/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is editsubject.
 *
 * The Initial Developer of the Original Code is
 * Takashi Hirata.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */

var editsubject = {

	//==================================
	// initialization
	onLoad: function(){
		this.initialized = true;
		this.strings = document.getElementById("editsubject-strings");
		this.oPrompt = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
							.getService(Components.interfaces.nsIPromptService);
		this.prefs = Components.classes["@mozilla.org/preferences-service;1"]
							.getService(Components.interfaces.nsIPrefBranch);
							//.getService(Components.interfaces.nsIPrefService).getBranch("");
		this.setupCharsetList(this.params.aCharsetList);

		// show context-editsubject only if single message is selected
		document.getElementById("mailContext").addEventListener(
			"popupshowing",
			function(e){
				document.getElementById("context-editsubject").hidden = (gFolderDisplay.selectedCount != 1);
			},
			false);
	},

	//==================================
	// internal parameters
	params : {
		strNewSubject		: null,
		strCharset			: null,
		nSelectedCharset	: 0,
		aCharsetList		: new Array(),
		boDirectCopy		: true,		// can't be false currently
		boDeleteOrg			: false,	// false for debug and test
		boDeleteTmp			: false,	// false for debug, valid only if DirectCopy=false
		boReplaceMode		: false,	// false for debug, valid only if DirectCopy=false
		dummy				: null
	},

	//######################################################
	// SECTION: event handlers (main function)
	//######################################################
	onMenuItemCommand: function(e){
		var uriMsg = gFolderDisplay.selectedMessageUris[0];
		var oMsgHdr = messenger.msgHdrFromURI(uriMsg);
		var oFolder = oMsgHdr.folder.QueryInterface(Components.interfaces.nsIMsgFolder);
		//var oFolder = gFolderDisplay.displayedFolder.QueryInterface(Components.interfaces.nsIMsgFolder);

		// process only local, real folder's message
		if ((oFolder.flags & 0x0020) // originally MSG_FOLDER_FLAG_VIRTUAL, but TB3.0 seems not to define this
			|| oFolder.server.type == "imap" || oFolder.server.type == "nntp")
		{
			this.alertWithLocale("illegalFolder", "FolderType: " + oFolder.server.type);
			return;
		}

		var xparams = this.params;
		xparams.boDeleteOrg = this.prefs.getBoolPref("extensions.editsubject.delete_original");
		xparams.strNewSubject = oMsgHdr.mime2DecodedSubject;
		xparams.strCharset = this.prefs.getCharPref("extensions.editsubject.charset");
		if(xparams.strCharset == ""){
			xparams.strCharset = xparams.aCharsetList[0].value;
		}
		// find matching charset, and set default selected one
		xparams.nSelectedCharset = -1;
		for(var i = 0 ; i < xparams.aCharsetList.length ; i++){
			if(xparams.strCharset == xparams.aCharsetList[i].value){
				xparams.nSelectedCharset = i;
				break;
			}
		}
		// if not found, add current charset
		// this is caused if user edited the pref by hand
		// and may be a reason of multibyte and mime encoding failure,
		// but useful for unlocalized user who doesn't use send/view encoding
		if(xparams.nSelectedCharset == -1){
			xparams.aCharsetList.push({ value : xparams.strCharset, label : xparams.strCharset });
			xparams.nSelectedCharset = xparams.aCharsetList.length - 1;
		}

		if(!this.openModalDialog("chrome://editsubject/content/newsubject_dialog.xul", "newsubject_dialog", xparams)){
			return;
		}

		// ======== exporting or DirectCopy ========
		var oStreamListner = this.getStreamListner();
		var oDest;
		if(xparams.boDirectCopy){
			var oDest = oFolder.QueryInterface(Components.interfaces.nsIMsgLocalMailFolder);
			oStreamListner.init(oDest, xparams.strNewSubject, xparams.strCharset, false);
		}else{
			// currently never used
			var oTempFile = this.createTempFile("editsubject.eml");
			// maybe last argument will be 'true', if importing does not need "From - xxxxxxx"
			oStreamListner.init(oTempFile, xparams.strNewSubject, xparams.strCharset, true);
		}

		messenger.messageServiceFromURI(uriMsg)
			.QueryInterface(Components.interfaces.nsIMsgMessageService)
			.streamMessage(uriMsg, oStreamListner, msgWindow, null, false, null);
		//@ may we as well use convertData? i know we shouldn't use additionalHeader
		// nsIURI streamMessage ( char* messageURI , nsISupports consumer ,
		//                        nsIMsgWindow msgWindow , nsIUrlListener urlListener ,
		//                        PRBool convertData , char* additionalHeader ) 

		if(!oStreamListner.getResult()){
			this.alertWithLocale("failedToExport", oTempFile.path);
			return;
		}

		// ======== importing ========
		// currently never used is code below inside if(!boDirectCopy){}
		// nsIMsgFolder.copyFileMessage() and nsIMsgCopyService.CopyFileMessage() complain
		// that they can't convert javascript argument. what the hell is this ??
		// I can't understand what's wrong within my code below all about
		// Thunderbird *MUST* DO RELEASE RELIABLE DOCUMENTATIONS !!
		//----
		if(!xparams.boDirectCopy){
			var oCopyListener = this.getCopyListner();
			oCopyListener.init();
			Components.classes['@mozilla.org/messenger/messagecopyservice;1']
				.createInstance(Components.interfaces.nsIMsgCopyService)
				.CopyFileMessage(oTempFile, oFolder, xparams.boReplaceMode ? oMsgHdr : null,
									false, oMsgHdr.flags, oCopyListener, msgWindow);
			// void CopyFileMessage ( nsIFile file , nsIMsgFolder dstFolder , nsIMsgDBHdr msgToReplace ,
			//                        PRBool isDraftOrTemplate , PRUint32 msgFlags ,
			//                        nsIMsgCopyServiceListener listener , nsIMsgWindow msgWindow ) 
			//----
			oFolder.copyFileMessage(oTempFile, xparams.boReplaceMode ? oMsgHdr : null, false,
									oMsgHdr.flags, msgWindow, oCopyListener);
			// void copyFileMessage ( nsIFile file , nsIMsgDBHdr msgToReplace , PRBool isDraft ,
			//                        PRUint32 newMsgFlags , nsIMsgWindow msgWindow ,
			//                        nsIMsgCopyServiceListener listener ) 

			if(!oCopyListener.getResult()){
				this.alertWithLocale("failedToImport", oTempFile.path);
				return;
			}
			if(xparams.boDeleteTmp){ oTempFile.remove(false); }
		}

		if((!xparams.boReplaceMode) && xparams.boDeleteOrg){ goDoCommand('cmd_delete'); }
		oFolder.updateFolder(msgWindow);
		oFolder.updateSummaryTotals(true);
		if(!xparams.boDirectCopy){
			gDBView.selectMsgByKey(oCopyListener.getKey());
		}
	},

	//######################################################
	// SECTION: importing funcitons
	// currently never used
	//######################################################
	getCopyListner: function(){
		var hsCopyServiceListener = {
			init : function(){
				this.boSuccess = true;
				this.nKey = -1;
			},
			// implements some @mozilla interface
			QueryInterface : function(iid)	{
				if (iid.equals(Components.interfaces.nsIMsgCopyServiceListener)
					|| iid.equals(Components.interfaces.nsISupports)
				){
					return this;
				}
				throw Components.results.NS_NOINTERFACE;
				return 0;
			},
			OnStartCopy	: function(){ },
			OnProgress	: function(progress, progressMax){ },
			OnStopCopy	: function(status){ },

			SetMessageKey: function(key){
				//Components.classes["@mozilla.org/consoleservice;1"].
				//	getService(Components.interfaces.nsIConsoleService).
				//	logStringMessage('SetMessageKey: key=' + key);
				this.nKey = key;
			},
			//GetMessageId : function(messageId){
			//	messageId = "<44428706.7010600@up-cat.net>";
			//},

			// returns true if successed
			getResult : function(){
				return this.boSuccess;
			},
			getKey : function(){
				return this.nKey;
			}
		};
		return hsCopyServiceListener;
	},

	//######################################################
	// SECTION: exporting funcitons
	//######################################################
	getStreamListner: function(){
		var hsStreamListner = {
			// should be called before each use
			//		oDest			: nsIFile or nsIMsgLocalMailFolder to be output
			//		strNewSubject	: new contents of "Subject: " in unicode
			//		boRemovePreFrom	: remove TB own header "From - DOW Mon dom hh:mm::ss yyyy, output may be generic EML
			// note: if oDest is a nsIMsgLocalMailFolder, boRemovePreFrom is ignored.
			init : function(oDest, strNewSubject, strCharset, boRemovePreFrom){
				this.cleanup();

				// setup new subject : 1. convert to local-multibyte first, 2. mime-encode the local-multibyte
				strNewSubject = "Subject: " + strNewSubject;
				var oUnicodeConverter = Components.classes['@mozilla.org/intl/scriptableunicodeconverter']
										.createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
				oUnicodeConverter.charset = strCharset;
				strNewSubject = oUnicodeConverter.ConvertFromUnicode(strNewSubject)
				var oMime = Components.classes["@mozilla.org/messenger/mimeconverter;1"]
										.getService(Components.interfaces.nsIMimeConverter);
				// how do we use parameter "fieldnamelen" ?
				// currently use 7, which is length of "Subject" in ascii()
				strNewSubject = oMime.encodeMimePartIIStr(strNewSubject, false, strCharset, 7, 72);

				this.oDest			= oDest;
				this.strNewSubject	= strNewSubject;
				if(oDest instanceof Components.interfaces.nsIMsgLocalMailFolder){
					this.fnOutputHandler = this.outputToFolder;
					this.boRemovePreFrom = false;
				}else{
					this.fnOutputHandler = this.outputToFile;
					this.boRemovePreFrom = boRemovePreFrom ? true : false;
				}

				this.oInStream = Components.classes['@mozilla.org/binaryinputstream;1']
									.createInstance(Components.interfaces.nsIBinaryInputStream);
				this.boSuccess		= true;
				this.initialized	= true;
			},
			// returns true if successed
			getResult : function(){
				return this.boSuccess;
			},
			cleanup : function(){
				this.oDest			= null;
				this.strNewSubject	= null;
				this.boRemovePreFrom = false;
				this.boFoundPreFrom	= false;		// already found pre TB From
				this.nSubjectState	= 0;			// state of processing "Subject: "
				this.oInStream		= null;			// input stream object
				this.initialized	= false;
				this.strLineDelim	= "\n";
				this.oDataPool		= "";			// only for DirectCopy mode
				this.fnInputStreamHandler = this.headerHandler;
				this.fnOutputHandler = null;
			},

			// implements some @mozilla interface
			QueryInterface : function(iid){
				if(iid.equals(Components.interfaces.nsIStreamListener)
					|| iid.equals(Components.interfaces.nsISupports)
					//|| iid.equals(Components.interfaces.nsIMsgHeaderSink)
				){
					return this;
				}
				throw Components.results.NS_NOINTERFACE;
				return 0;
			},
			onStartRequest : function(request, context){
			},
			onStopRequest : function(request, context, statusCode){
				this.oInStream.close();
				if(this.oDataPool.length > 0){
					this.oDest.addMessage(this.oDataPool);
				}
				this.cleanup();
			},
			onDataAvailable : function(request, context, inputStream, offset, count){
				//editsubject.alert("VF: count=" + count);
				this.oInStream.setInputStream(inputStream);
				while((count = this.fnInputStreamHandler(this.oInStream, count)) != 0){ 
					; 
				}
			},

			// process inputStream till end of header
			// when reach to end of header, substitute fnInputStreamHandler to process body
			// byte-by-byte operation
			// "\n\r" and "\r" pattern of newline is not supported
			headerHandler : function(oInputStream, nCount){
				var strLine = "";
				while(nCount--){
					var nByte = oInputStream.readBytes(1);
					strLine += nByte;
					if(nByte == "\r"){ this.strLineDelim = "\r\n"; };
					if(nByte == "\n"){
						// now a line is constructed
						if(strLine.search(/^[\r\n]*$/) != -1){
							// detected end of header, output new subject
							this.fnOutputHandler(this.strNewSubject + this.strLineDelim);
							this.fnOutputHandler(strLine);
							this.strNewSubject = null;
							// substitute line handler
							this.fnInputStreamHandler = this.bodyHandler;
							return nCount;
						}
						if(this.checkHeaderLine(strLine)){
							this.fnOutputHandler(strLine);
						}
						strLine = "";
					}
				}
				return nCount;
			},
			// process inputStream by large unit, for message body
			bodyHandler : function(oInputStream, nCount){
				var oData = oInputStream.readBytes(nCount);
				this.fnOutputHandler(oData);
				return nCount - oData.length;
			},
			// check single line of header
			// return true if the strLine should be output
			checkHeaderLine : function(strLine){
				if(strLine.search(/^Subject:/) != -1){
					this.nSubjectState = 1;		// detected "Subject:" first line, remove it, and move state
					return false;
				}
				if(this.nSubjectState == 1){
					if(strLine.search(/^[ \t]/) != -1){
						return false;	// detect "Subject:" subsequential line, remove it
					}
					this.nSubjectState = 2;
				}
				if(strLine.search(/^From .+/) != -1){
					if(this.boFoundPreFrom){
						// found twice
						editsubject.alertWithLocale("illegalFromFound");
						this.boSuccess = false;
					}else if(this.boRemovePreFrom){
						return false;	// remove "From - DOW Mon dom hh:mm::ss yyyy
					}
					this.boFoundPreFrom = true;
				}
				return true;
			},
			outputToFolder : function(strData){
				this.oDataPool += strData;
			},
			outputToFile : function(strData){
				editsubject.writeFile(this.oDest, strData, true);
			}
		};
		return hsStreamListner;
	},

	//######################################################
	// SECTION: utility functions
	//######################################################

	// charset can be such as "UTF-8"
	writeFile: function(oFile, oData, boAppend, strCharset){
		var PR_RDONLY	= 0x01;	// Open for reading only.
		var PR_WRONLY	= 0x02;	// Open for writing only.
		var PR_RDWR		= 0x04;	// Open for reading and writing.
		var PR_CREATE_FILE = 0x08;	// If the file does not exist, the file is created. 
									//If the file exists, this flag has no effect.
		var PR_APPEND	= 0x10;	// The file pointer is set to the end of the file prior to each write.
		var PR_TRUNCATE	= 0x20;	// If the file exists, its length is truncated to 0.
		var PR_SYNC		= 0x40;	// If set, each write will wait 
								// for both the file data and file status to be physically updated.
		var PR_EXCL		= 0x80;	// Used with PR_CREATE_FILE; if the file does not exist, the file is created.
								// If the file already exists, no action and NULL is returned 

		var oOutFileStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
								.createInstance(Components.interfaces.nsIFileOutputStream);
		var nFlags = PR_WRONLY | PR_CREATE_FILE;
		nFlags |= boAppend ? PR_APPEND : PR_TRUNCATE;
		oOutFileStream.init(oFile, nFlags, -1, 0); // ( nsIFile, ioFlags, perm=0664, [reserved])

		if(strCharset){
			var oConvStream = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
								.createInstance(Components.interfaces.nsIConverterOutputStream);
			oConvStream.init(oOutFileStream, strCharset, 0, "?".charCodeAt(0)); // (outStream, charset, bufferSize=nobuffer, replacementCharacter);
			if(oData){
				oConvStream.writeString(data);
			}
			oConvStream.close();
		}else{
			if(oData){
				oOutFileStream.write(oData, oData.length);
			}
		}
		oOutFileStream.close();
	},
	createTempFile: function(strCandidateName){
		var oTempFile = Components.classes["@mozilla.org/file/directory_service;1"]
						.getService(Components.interfaces.nsIProperties)
						.get("TmpD", Components.interfaces.nsIFile);
		oTempFile.append(strCandidateName);
		oTempFile.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0664);
		return oTempFile;
	},
	openModalDialog : function(urlDialog, strName, oArg){
		var hsParam = { in: oArg, out: null };
		window.openDialog(urlDialog, strName, "titlebar,close,chrome,dependent=no,dialog,modal,centerscreen", hsParam);
		return hsParam.out;
	},
	alert : function(strMessageText){
		this.oPrompt.alert(window, this.strings.getString("editsubjectTitle"), strMessageText);
	},
	alertWithLocale : function(strMessageId, strAdditionalMessage){
		if(strAdditionalMessage){
			this.alert(this.strings.getString(strMessageId) + strAdditionalMessage);
		}else{
			this.alert(this.strings.getString(strMessageId));
		}
	},
	setupCharsetList : function(aCharsetList){
		aCharsetList.push({ value : this.prefs.getComplexValue("mailnews.send_default_charset", Components.interfaces.nsIPrefLocalizedString).data });
		aCharsetList.push({ value : this.prefs.getComplexValue("mailnews.view_default_charset", Components.interfaces.nsIPrefLocalizedString).data });
		aCharsetList.push({ value : "ISO-2022-JP" });	// [Japanese]
		aCharsetList.push({ value : "ISO-8859-1" });	// [Western] old Western standard, similar to US-ASCII but support other languages
		aCharsetList.push({ value : "ISO-8859-15" });	// [Western(2)] new Western standard, coverage improved - such as Euro glyph, French "oe"
		aCharsetList.push({ value : "ISO-8859-6" });	// [Arabic]
		aCharsetList.push({ value : "ARMSCII-8" });		// [Armenian]
		aCharsetList.push({ value : "GEOSTD8" });		// [Georgian]
		aCharsetList.push({ value : "ISO-8859-13" });	// [Baltic]
		aCharsetList.push({ value : "ISO-8859-14" });	// [Celtic]
		aCharsetList.push({ value : "ISO-8859-2" });	// [Central European]
		aCharsetList.push({ value : "GB2312" });		// [Chinese Simplified] old Chinese standard (EUC-CN)
		aCharsetList.push({ value : "GB18030" });		// [Chinese Simplified(2)] new Chinese standard, extends GB2312
		aCharsetList.push({ value : "Big5" });			// [Chinese Traditional]
		aCharsetList.push({ value : "KOI8-R" });		// [Cyrillic] i don't know the difference of these Cyrillics
		aCharsetList.push({ value : "Windows-1251" });	// [Cyrillic(2)]
		aCharsetList.push({ value : "KOI8-U" });		// [Cyrillic/Ukrainian]
		aCharsetList.push({ value : "ISO-8859-7" });	// [Greek]
		aCharsetList.push({ value : "ISO-8859-8-I" });	// [Hebrew] i don't know the difference of these Hebrews
		aCharsetList.push({ value : "Windows-1255" });	// [Hebrew(2)]
		aCharsetList.push({ value : "EUC-KR" });		// [Korean]
		aCharsetList.push({ value : "ISO-8859-10" });	// [Nordic] North Eruropean
		aCharsetList.push({ value : "ISO-8859-3" });	// [South European]
		aCharsetList.push({ value : "TIS-620" });		// [Thai]
		aCharsetList.push({ value : "ISO-8859-9" });	// [Turkish]
		aCharsetList.push({ value : "UTF-8" });			// [Unicode]
		aCharsetList.push({ value : "VISCII" });		// [Vietnamese]
		// setup labels
		aCharsetList[0].label = editsubject.strings.getString("charsetDescription_FollowSend");
		aCharsetList[0].label += " (" + aCharsetList[0].value + ")";
		aCharsetList[1].label = editsubject.strings.getString("charsetDescription_FollowView");
		aCharsetList[1].label += " (" + aCharsetList[1].value + ")";
		for(var i = 2 ; i < aCharsetList.length ; i++){
			aCharsetList[i].label = editsubject.strings.getString("charsetDescription_" + aCharsetList[i].value);
			aCharsetList[i].label += " (" + aCharsetList[i].value + ")";
		}
	}
};

window.addEventListener("load", function(e) { editsubject.onLoad(e); }, false);
